(function (factory) {
    if (typeof module === "object" && typeof module.exports === "object") {
        var v = factory(require, exports);
        if (v !== undefined) module.exports = v;
    }
    else if (typeof define === "function" && define.amd) {
        define(["require", "exports", "rxjs", "./observedObjectsSymbol"], factory);
    }
})(function (require, exports) {
    "use strict";
    Object.defineProperty(exports, "__esModule", { value: true });
    var _a;
    const rxjs_1 = require("rxjs");
    const observedObjectsSymbol_1 = require("./observedObjectsSymbol");
    class ReusableSubject extends rxjs_1.Subject {
        unsubscribe() {
            this.observers = [];
        }
    }
    class _ObservableObject {
        constructor(from = {}, optHandlers = {}) {
            this[_a] = {};
            let setCustomHandler;
            let getCustomHandler;
            if (optHandlers) {
                if (optHandlers.set) {
                    setCustomHandler = optHandlers.set;
                    delete optHandlers.set;
                }
                if (optHandlers.get) {
                    getCustomHandler = optHandlers.get;
                }
            }
            const handlers = Object.assign(optHandlers, {
                set: (obj, prop, value, receiver, ...args) => {
                    let notificationChain;
                    if (typeof value === "object" && !Array.isArray(value)) {
                        notificationChain = Object.assign({
                            [prop]: value,
                        }, buildNotificationChain(value, prop));
                        if (setCustomHandler) {
                            let setResult = setCustomHandler(obj, prop, value, receiver);
                            if (setResult === false) {
                                return setResult;
                            }
                        }
                        obj[prop] = new Proxy(value, Object.assign(handlers, {
                            set: bindLast(handlers.set, [...args, prop]),
                        }));
                    }
                    else {
                        if (obj[prop] === value) {
                            return true;
                        }
                        if (setCustomHandler) {
                            let setResult = setCustomHandler(obj, prop, value, receiver);
                            if (setResult === false) {
                                return setResult;
                            }
                        }
                        obj[prop] = value;
                        const elementKey = args.length && args[0].length ? [...args[0], prop].join(".") : prop;
                        notificationChain = {
                            [elementKey]: value
                        };
                    }
                    Object.keys(notificationChain).forEach((keyPath) => {
                        const value = notificationChain[keyPath];
                        if (this[observedObjectsSymbol_1.observedObjects][keyPath]) {
                            this[observedObjectsSymbol_1.observedObjects][keyPath].next(value);
                        }
                    });
                    return true;
                },
                get: bindLast((target, prop, receiver, customGetter) => {
                    if (customGetter !== undefined && !(prop in _ObservableObject.prototype)) {
                        return customGetter(target, prop, receiver);
                    }
                    return Reflect.get(target, prop, receiver);
                }, getCustomHandler)
            });
            return new Proxy(Object.assign(this, buildInitialProxyChain(from, handlers)), handlers);
        }
        observe(prop) {
            if (!this[observedObjectsSymbol_1.observedObjects][prop]) {
                this[observedObjectsSymbol_1.observedObjects][prop] = new ReusableSubject();
            }
            return this[observedObjectsSymbol_1.observedObjects][prop];
        }
        unsubscribeAll(subscriptions) {
            subscriptions.forEach(sub => sub.unsubscribe());
        }
        snapshot() {
            const snapshot = Object.assign({}, this);
            delete snapshot[observedObjectsSymbol_1.observedObjects];
            return snapshot;
        }
    }
    _a = observedObjectsSymbol_1.observedObjects;
    exports.ObservableObject = _ObservableObject;
    function buildInitialProxyChain(sourceObject, handlers, ...args) {
        let chain = {};
        for (const prop in sourceObject) {
            if (typeof sourceObject[prop] === "object" && !Array.isArray(sourceObject[prop])) {
                chain[prop] = buildInitialProxyChain(sourceObject[prop], Object.assign(handlers, {
                    set: bindLast(handlers.set, [...args, prop])
                }), ...args, prop);
            }
            else {
                chain[prop] = sourceObject[prop];
            }
        }
        return new Proxy(chain, handlers);
    }
    function buildNotificationChain(source, ...args) {
        let chain = {};
        for (const prop in source) {
            chain[[...args, prop].join(".")] = source[prop];
            if (typeof source[prop] === "object" && !Array.isArray(source[prop])) {
                Object.assign(chain, buildNotificationChain(source[prop], ...args, prop));
            }
        }
        return chain;
    }
    function bindLast(fn, ...boundArgs) {
        return (...args) => fn(...args, ...boundArgs);
    }
});
